from setuptools import setup

with open("README.md", "r") as fh:
    readme = fh.read()

setup(
    name="iso2mesh",
    packages=["iso2mesh"],
    version="0.4.2",
    license="GPLv3+",
    description="One-liner 3D Surface and Tetrahedral Mesh Generation Toolbox",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Qianqian Fang",
    author_email="fangqq@gmail.com",
    maintainer="Qianqian Fang",
    url="https://github.com/NeuroJSON/pyiso2mesh",
    keywords=[
        "Iso2Mesh",
        "Brain2Mesh",
        "JSONLab",
        "EasyH5",
        "Mesh generation",
        "Finite element method",
        "FEM",
        "FEA",
        "MATLAB",
        "Mesh-based Monte Carlo",
        "CGAL",
        "Tetgen",
        "CSG",
        "Boolean",
        "Surface mesh",
        "Tetrahedral mesh",
        "Modeling",
        "Multiphysics",
        "Image processing",
        "NeuroJSON",
        "Binary JSON",
        "JSON",
        "Data format",
    ],
    platforms="any",
    install_requires=["numpy>=1.8.0", "matplotlib", "jdata>=0.7.1", "bjdata"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
